from meteor_reasoner.materialization.join_util import *
from meteor_reasoner.materialization.apply import *


def ground_unary_literal(literal, context, D, D_index=None):
    """
    This function is a generator, which generates entity, interval and the context.
    Args:
        literal (a literal instance):
        context (a dictionary): store the mapping of  variables and constants
        D (a dictionary object): a global object storing all facts.

    Returns:
        entity, interval, and the context

    """
    predicate = literal.get_predicate()
    entity = copy.deepcopy(literal.get_entity())

    if predicate in D:
        if len(entity) == 1 and entity[0].name == "nan":
            if predicate in D:
                yield entity, D[predicate][entity], dict()


        elif not contain_variable(entity):
            if predicate in D and entity in D[predicate]:
                yield entity, D[predicate][entity], dict()

        elif not contain_variable_after_replace(entity, context):
            replaced_entity = []
            for term in entity:
                if term.type == "variable":
                    term.type = "constant"
                    term.name = context[term.name]
                    replaced_entity.append(term)
                else:
                    replaced_entity.append(term)

            if predicate in D and tuple(replaced_entity) in D[predicate]:
                yield tuple(replaced_entity), D[predicate][tuple(replaced_entity)], dict()

        else:
            if D_index is not None:
                index_str = []
                for i, term in enumerate(entity):
                    if term.type == "constant":
                        index_str.append(str(i) + "@" + term.name)
                    else:
                        if term.name in context:
                            index_str.append(str(i)+"@"+context[term.name])

                index_str = "||".join(index_str)
                if len(index_str) == 0:
                    for constant_entity, intervals in copy.deepcopy(D[predicate]).items():
                        tmp_context = dict()
                        for term1, term2 in zip(entity, constant_entity):
                            if term1.type == "constant" and term1.name != term2.name:
                                break
                            elif term1.type == "variable" and term1.name in context and context[
                                term1.name] != term2.name:
                                break
                            elif term1.type == "variable" and term1.name in tmp_context and tmp_context[
                                term1.name] != term2.name:
                                break
                            else:
                                if term1.type == "variable":
                                    tmp_context[term1.name] = term2.name
                        else:
                            yield constant_entity, D[predicate][constant_entity], tmp_context
                else:
                    if index_str in D_index[predicate]:
                        for constant_entity in D_index[predicate][index_str]:
                            tmp_context = dict()
                            for term1, term2 in zip(entity, constant_entity):
                                if term1.type == "constant" and term1.name != term2.name:
                                    break
                                elif term1.type == "variable" and term1.name in context and context[term1.name] != term2.name:
                                    break
                                elif term1.type == "variable" and term1.name in tmp_context and tmp_context[term1.name] != term2.name:
                                    break
                                else:
                                    if term1.type == "variable":
                                        tmp_context[term1.name] = term2.name
                            else:
                                yield constant_entity, D[predicate][constant_entity], tmp_context

            else:
                for tmp_entity in D[predicate]:
                    tmp_context = dict()
                    for term1, term2 in zip(entity, tmp_entity):
                        if term1.type == "constant" and term1.name != term2.name:
                            break
                        elif term1.type == "variable" and term1.name in context and context[term1.name] != term2.name:
                            break
                        elif term1.type == "variable" and term1.name in tmp_context and tmp_context[term1.name] != term2.name:
                            break
                        else:
                            if term1.type == "variable":
                                tmp_context[term1.name] = term2.name
                    else:
                        yield tmp_entity, D[predicate][tmp_entity], tmp_context



from meteor_reasoner.materialization.join import *
from meteor_reasoner.materialization.coalesce import *
from meteor_reasoner.materialization.index_build import *
from meteor_reasoner.utils.loader import load_program, load_dataset
from meteor_reasoner.utils.operate_dataset import *
import time


def materialize(D, rules, common_fragment=None, K=100):
    """
    The function implements the materialization operation.
    Args:
        D (dictionary object):
        rules (list of Rule instances):
        K (int): the nunber of steps for exectuting the materialization. By default=100.
    Returns:
        The
    """
    k = 0
    while k < K:
        coalescing_d(D)
        D_index = defaultdict(lambda: defaultdict(list))
        build_index(D, D_index)
        k += 1
        if k == 1:
            delta_old = D
            delta_new = defaultdict(lambda: defaultdict(list))
        else:
            coalescing_d(delta_new)
            delta_old = delta_new
            delta_new = defaultdict(lambda: defaultdict(list))

        for i, rule in enumerate(rules):
            join(rule, D, delta_old, delta_new, common_fragment, D_index)

        if not delta_new:
            return True, delta_new

        if common_fragment is None:
            #add to D later
            for predicate in delta_new:
                for entity in delta_new[predicate]:
                    if predicate not in D:
                        D[predicate][entity] = delta_new[predicate][entity]
                    elif predicate in D and entity not in D[predicate]:
                        D[predicate][entity] = delta_new[predicate][entity]
                    elif predicate in D and entity in D[predicate]:
                        D[predicate][entity] += delta_new[predicate][entity]

    if common_fragment is None:
         coalescing_d(D)

    return False, delta_new


if __name__ == "__main__":
    raw_data = ["P(mike,mike)@1"]
    raw_program = ["P(X,X):- Diamondminus[0,1]P(X,X)"]
    D = load_dataset(raw_data)
    Program = load_program(raw_program)
    start_time = time.time()
    flag = materialize(D, Program, K=1)
    end_time = time.time()
    print("The mat time:", end_time-start_time)

    #print_dataset(D)











